/*
 * Copyright (c) 2000-2006 CNRS/LAAS
 *
 * Anthony Mallet - Wed Aug  9 2000
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/*	$LAAS: camera.c,v 1.5 2006/09/29 17:45:30 mallet Exp $ */

/*
 * plugin arguments -----------------------------------------------------
 *
 * bank = <a|b|c|...>	image poster
 * side = <left|right>	image number
 * width = <int>	size of image
 * height = <int>
 * quality = <int 0-99>	jpeg quality
 * delay = <int>	delay (ms*10) between transmission
 */

#include <portLib.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <taskLib.h>

#include <jpeglib.h>
#include <jerror.h>
#include "jmem.h"

#include <posterLib.h>
#include <h2timeLib.h>

#include <pom/pomStruct.h>
#include <camera/cameraClient.h>

extern STATUS	webError(char *text, int n, int fd);

static int	jpgCompress(unsigned char *dst, unsigned char *buf,
			    int width, int height, int quality, int color);
static void	reduce(unsigned char *src, int ow, int oh, int rw, int rh,
			int color);


/*
 * ----------------------------------------------------------------------
 */

void
webcamera(FILE *f, int argc, char **argv, char **argn)
{
#define SEP	"jpgimage123imagejpg"

	H2TIME date, last;
	unsigned char *pixels, *compress;
	unsigned char *source;
	int i, pWidth, pHeight, length, color;
	POSTER_ID posterId;
	CameraPoster *poster;
	STATUS status;
	char buf[512];
	struct timespec tv;

	int bank = 0;			/* plugin defaults */
	int image = CAMERA_LEFT;
	int width = 256;
	int height = 192;
	int quality = 30;
	int delay = -1;

	/* parse other args */
	for(i = 1; i < argc; i++) {
		if (argn[i] == NULL) continue;
		if (!strcmp(argn[i], "bank")) bank = tolower(argv[i][0]) - 'a';
		else if (!strcmp(argn[i], "image")) {
			if (!strcmp(argv[i], "mono")) image = CAMERA_MONO;
			else if (!strcmp(argv[i], "left")) image = CAMERA_LEFT;
			else if (!strcmp(argv[i], "right")) image = CAMERA_RIGHT;
			else {
				webError("bad image", 404, fileno(f));
				return;
			}
		}
		else if (!strcmp(argn[i], "width")) width = atoi(argv[i]);
		else if (!strcmp(argn[i], "height")) height = atoi(argv[i]);
		else if (!strcmp(argn[i], "quality")) quality = atoi(argv[i]);
		else if (!strcmp(argn[i], "delay")) delay = atoi(argv[i]);
	}

	/* look for poster */
	sprintf(buf, CAMERA_POSTER_NAME, bank);
	if (posterFind(buf, &posterId) != OK) {
		webError("no poster", 404, fileno(f));
		return;
	}

	poster = posterAddr(posterId);

	if (posterTake(posterId, POSTER_READ) != OK) {
		webError("no poster take", 404, fileno(f));
		return;
	}

	/* check size */
	if (width > poster->width) width = poster->width;
	if (height > poster->height) height = poster->height;

	/* create internal buffer for image compression */
	pWidth = poster->width;
	pHeight = poster->height;
	color = poster->bayer & CAMERA_BAYER_COLOR;
	pixels = malloc(pWidth*pHeight * (color?4:1));
	if (pixels == NULL) {
		posterGive(posterId);
		webError("no memory", 500, fileno(f));
		return;
	}

	compress = malloc(pWidth*pHeight * (color?3:1));
	/* ! 24bits/pixel in color mode */
	if (compress == NULL) {
		posterGive(posterId);
		free(pixels);
		webError("no memory", 500, fileno(f));
		return;
	}

	/* compute image address */
	switch(image) {
	case CAMERA_MONO:
		if (poster->nImages != 1) {
			posterGive(posterId);
			webError("no image", 404, fileno(f));
			free(pixels); free(compress);
			return;
		}
		source = poster->image.mono.pixel;
		break;

	case CAMERA_LEFT:
	case CAMERA_RIGHT:
		if (poster->nImages != 2) {
			posterGive(posterId);
			webError("no image", 404, fileno(f));
			free(pixels); free(compress);
			return;
		}

		if (image == CAMERA_LEFT)
		  source = poster->image.stereo.pixel +
		    CAMERA_POSTER_LEFT*
		    poster->maxWidth*poster->maxHeight *
		    (color?4:1);
		else
		  source = poster->image.stereo.pixel +
		    CAMERA_POSTER_RIGHT*
		    poster->maxWidth*poster->maxHeight *
		    (color?4:1);
		break;

	default:
		posterGive(posterId);
		webError("no image", 404, fileno(f));
		free(pixels); free(compress);
		return;
	}

	posterGive(posterId);

	/* start transmission */
	/* setbuf(f, NULL); */

	fprintf(f, "HTTP/1.1 200 OK\nPragma: no-cache\n");
	fprintf(f, "Server: picoweb 1.0\n");
	if (delay != -1) {
		fprintf(f,
		    "Content-Type: multipart/x-mixed-replace;boundary=" SEP "\n\n");
		fprintf(f, "\n--" SEP "\n");
	} else {
		fprintf(f, "Connection: close\n");
	}
	fprintf(f, "Content-Type: image/jpeg\n");

	while(1) {

		if (delay != -1) {
			status = posterIoctl(posterId, FIO_GETDATE, (char *)&date);
			if (status == OK && date.ntick == last.ntick) {
				tv.tv_sec = delay / sysClkRateGet();
				tv.tv_nsec = (delay % sysClkRateGet()) * 1000000000/ sysClkRateGet();
				nanosleep(&tv, NULL);
				continue;
			}
			last = date;
		}

		if (posterTake(posterId, POSTER_READ) != OK) {
			webError("camera(take)", 503, fileno(f)); break;
		}
		memcpy(pixels, source, pWidth*pHeight *(color?4:1));
		posterGive(posterId);

		if (pWidth != width || pHeight != height)
		  reduce(pixels, pWidth, pHeight, width, height, color);

		length = jpgCompress(compress, pixels, width, height,
				     quality, color);
		if (length < 0) {
			webError("camera(compress)", 500, fileno(f)); break;
		}

		fprintf(f, "Content-Length: %d\n\n", length);
		if (fwrite(compress, length, 1, f) != 1) {
			fprintf(stderr, "short write\n");
			break;
		}

		if (delay == -1) {
			/* only send one image */
			break;
		}

		tv.tv_sec = delay / sysClkRateGet();
		tv.tv_nsec = (delay % sysClkRateGet()) * (1000000000/ sysClkRateGet());
		nanosleep(&tv, NULL);
		fprintf(f, "\n--" SEP "\nContent-Type: image/jpeg\n");
	}

	free(pixels);
	free(compress);
}

static int
jpgCompress(unsigned char *dst, unsigned char *buf,
	    int width, int height, int quality, int color)
{
	struct jpeg_compress_struct cinfo;
	struct jpeg_error_mgr jerr;
	JSAMPROW row_pointer[1];
	char *d, *s;
	int i, length;

	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_compress(&cinfo);

	cinfo.image_width = width;
	cinfo.image_height = height;
	if (color) {
	  cinfo.input_components = 3;
	  cinfo.in_color_space = JCS_RGB;
	  row_pointer[0] = malloc(3*width);
	} else {
	  cinfo.input_components = 1;
	  cinfo.in_color_space = JCS_GRAYSCALE;
	}
	jpeg_mem_dest(&cinfo, dst, cinfo.input_components*width*height);

	jpeg_set_defaults(&cinfo);
	jpeg_set_quality(&cinfo, quality, FALSE);
	jpeg_start_compress(&cinfo, TRUE);
	while (cinfo.next_scanline < cinfo.image_height) {
		/* jpeg_write_scanlines expects an array of pointers
		 * to scanlines.  Here the array is only one element
		 * long, but you could pass more than one scanline at
		 * a time if that's more convenient.
		 */
	  if (color) {
	    /* copie ligne RGBA -> RGB */
	    s = buf + cinfo.next_scanline * 4*width;
	    d = row_pointer[0];
	    for(i=0; i<width; i++) {
	      *(d++) = *(s++);
	      *(d++) = *(s++);
	      *(d++) = *(s++);
	      s++;
	    }
	  } else {
	    /* Adresse de la ligne */
	    row_pointer[0] = buf + cinfo.next_scanline * width;
	  }
	  jpeg_write_scanlines(&cinfo, row_pointer, 1);
	}
	jpeg_finish_compress(&cinfo);
	length = jpeg_mem_size(&cinfo);
	jpeg_destroy_compress(&cinfo);
	if (color) free(row_pointer[0]);
	return length;
}

static void
reduce(unsigned char *src, int ow, int oh, int rw, int rh, int color)
{
	unsigned char *dst = src;
	double hrest, vrest, herr, verr;
	int hstep, vstep, htot;
	int i, j;

	hstep = ow/rw;			vstep = oh/rh;
	hrest = (double)ow/rw - hstep;	vrest = (double)oh/rh - vstep;

	/* one line is already skipped by column scan */
	vstep--;

	verr = 0.0;
	for(i=0;i<rh;i++) {

		herr = 0.0;
		htot = 0;
		for(j = 0; j < rw; j++) {
		  if (!color) {
		    *dst = *src;
		    dst++;
		  } else {
		    *(uint32_t *)dst = *(uint32_t *)src;
		    dst+=4;
		  }
		  herr += hrest;
		  if (herr >= 1.0) {
		    herr -= 1.0;
		    src+= color?4:1;
		    htot++;
		  }
		  src += hstep * (color?4:1);
		  htot += hstep;
		}
		src+= (ow-htot) * (color?4:1);

		verr += vrest;
		if (verr >= 1.0) {
			verr -= 1.0;
			src+=ow * (color?4:1);
		}
		src += ow*vstep * (color?4:1);
	}
}
