/*
 * Copyright (c) 2006 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ident "$Id: autoload.c,v 1.10 2007/06/06 14:44:09 matthieu Exp $"

#include <dlfcn.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <portLib.h>
#include <posterLib.h>
#include <h2errorLib.h>

#include "picoweb.h"

static PICOWEB_MODULE *loadedModules = NULL;
static PICOWEB_MODULE *loadmodule(char*, char *);
static char *moduleweb(char *);
static char *moduleclient(char *);
static char *modulecgi(char *);
static char *moduleposterinit(char *);

CGIFUNCPTR
autoload(char *moduleName)
{
	char *base, *webbase, *symbol, *posterinit;
	char *errorMsg = NULL;
	CGIFUNCPTR fptr;
	INTFUNCPTR posterinitfunc;
	PICOWEB_MODULE *m;

	symbol = modulecgi(moduleName);
	if (symbol == NULL)
		return NULL;
	for (m = loadedModules; m != NULL; m = m->next) {
        	if (strcmp(moduleName,m->name) == 0 && (fptr = (CGIFUNCPTR)dlsym(m->handle, symbol)) != NULL)
			break;
	}
	if (m == NULL) {
		/* need to load the module */
		/* try <module>Web.so first */
		webbase = moduleweb(moduleName);
		m = loadmodule(moduleName, webbase);
		if (m == NULL) {
			errorMsg = strdup(dlerror());
			/* try lib<module>Client.so */
			base = moduleclient(moduleName);
			m = loadmodule(moduleName, base);
			if (m == NULL) {
				/* failed */
				fprintf(stderr, "picoweb: %s\n\t%s\n",
				    errorMsg, dlerror());
				free(errorMsg);
				free(symbol);
				free(webbase);
				free(base);
				return NULL;
			}
			free(errorMsg);
			free(webbase);
		} else {
			base = webbase;
		}
		if ((fptr = (CGIFUNCPTR)dlsym(m->handle, symbol)) == NULL) {
			fprintf(stderr, "picoweb: %s: %s\n", __func__,
			    dlerror());
			free(symbol);
			free(base);
			free(m);
			return NULL;
		}
		/* init poster client */
		posterinit = moduleposterinit(moduleName);
		if ((posterinitfunc = (INTFUNCPTR)dlsym(m->handle,
			    posterinit)) == NULL) {
			fprintf(stderr, "picoweb: %s: %s\n", __func__,
			    dlerror());
			free(posterinit);
			free(symbol);
			free(base);
			free(m);
			return NULL;
		}
		/* Call the init function */
		if (posterinitfunc() != OK) {
			fprintf(stderr, "picoweb: %s: ", posterinit);
			h2printErrno(errnoGet());
			/* A failure of the posterInit function is not 
			   fatal, as there are some modules (pom) for
			   which it will always fail. sigh */
		}
		free(posterinit);
		free(base);
		m->next = loadedModules;
		loadedModules = m;
	}
	free(symbol);
	return(fptr);
}

static PICOWEB_MODULE *
loadmodule(char *moduleName, char *basename)
{
	PICOWEB_MODULE *m;
	void *h;
	size_t len;

	h = dlopen(basename, RTLD_NOW);
	if (h == NULL) 
		return NULL;

	m = (PICOWEB_MODULE *)malloc(sizeof(PICOWEB_MODULE));
	if (m == NULL) {
		dlclose(h);
		return NULL;
	}
	m->next = NULL;
	m->handle = h;
	len = strlen(moduleName);
	m->name = malloc(len);
	strcpy(m->name, moduleName);
	return m;
}

static char *
moduleweb(char *moduleName)
{
	char *buf;
	size_t len;

	len = strlen(moduleName) + 7;
	buf = malloc(len);
	if (buf == NULL) {
		fprintf(stderr, "picoweb: %s: malloc failure\n", __func__);
		return NULL;
	}
	snprintf(buf, len, "%sWeb.so", moduleName);
	return buf;
}
	

static char *
moduleclient(char *moduleName)
{
	char *buf;
	size_t len;

	len = strlen(moduleName) + 13;
	buf = malloc(len);
	if (buf == NULL) {
		fprintf(stderr, "picoweb: %s: malloc failure\n", __func__);
		return NULL;
	}
	snprintf(buf, len, "lib%sClient.so", moduleName);
	return buf;
}

static char *
modulecgi(char *moduleName)
{
	char *buf;
	size_t len;

	len = strlen(moduleName) + 4;
	buf = malloc(len);
	if (buf == NULL) {
		fprintf(stderr, "picoweb: %s: malloc failure\n", __func__);
		return NULL;
	}
	snprintf(buf, len, "web%s", moduleName);
	return buf;
}

static char *
moduleposterinit(char *moduleName)
{
	char *buf;
	size_t len;

	len = strlen(moduleName) + 11;
	buf = malloc(len);
	if (buf == NULL) {
		fprintf(stderr, "picoweb: %s: malloc failure\n", __func__);
		return NULL;
	}
	snprintf(buf, len, "%sPosterInit", moduleName);
	return buf;
}

